/* midi - MIDI file library and Simple Software Synthesizer
 * Copyright (C) 2004 by Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <glib.h>
#include <stdio.h>
#include <string.h>

#define DEBUG_LEVEL 0
#include "midifile.h"

MIDIFile *read_midi_file(FILE *fp) {
	char id[5] = "....";
	unsigned long len;
	MIDIFile *mf;

	mf = g_new(MIDIFile, 1);
	mf->tracks = g_ptr_array_new();
	mf->header = NULL;

	while (!feof(fp)) {
		if (fread(id, 4, 1, fp) < 1) break;
		len = read_int32(fp);
		if (!strcmp(id, "MThd")) {
			g_assert(mf->header == NULL);
			mf->header = read_header_chunk(fp, len);
		}
		else if (!strcmp(id, "MTrk")) {
			Track *track = read_track(fp, len);
			g_ptr_array_add(mf->tracks, track);
		}
		else {
			debug(1, "Warning: ignoring track type \"%s\"\n", id);
			fseek(fp, len, SEEK_CUR);
		}
	}
	g_assert(mf->header != NULL);
	g_assert(mf->tracks->len == mf->header->numtracks);

	return mf;
}

Header *read_header_chunk(FILE *fp, int length) {
	Header *chunk = g_new(Header, 1);
	g_assert(length == 6);
	chunk->format = read_int16(fp);
	chunk->numtracks = read_int16(fp);
	chunk->division = read_int16(fp);

	g_assert(chunk->format <= 2);
	if (chunk->format == 0) g_assert(chunk->numtracks == 1);

	return chunk;
}

Track *read_track(FILE *fp, int length) {
	long start_pos;
	Track *track = g_new(Track, 1);
	Event *event = NULL;

	track->events = g_ptr_array_new();
	start_pos = ftell(fp);
	do {
		event = read_event(fp, event);
		g_ptr_array_add(track->events, event);
	} while (event->status != 0xff || event->pitch != 0x2f || event->velocity != 0x00);
	g_assert(ftell(fp) - start_pos == length);
	return track;
}

const char *note[] = {
	"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"
};
Event *read_event(FILE *fp, Event *last_event) {
	int delta_time, len;
	Event *event = g_new(Event, 1);

	delta_time = read_varint(fp);
	event->start = (last_event ? last_event->start : 0) + delta_time;
	event->status = getc(fp);
	event->data = NULL;
	
	if (event->status < 0x80) {
		g_assert(last_event != NULL);
		ungetc(event->status, fp);
		event->status = last_event->status;
	}

	if (event->status >= 0x80 && event->status <= 0x9f) {
		event->pitch = getc(fp);
		event->velocity = getc(fp);
		debug(3, "Note: @ %d status=0x%02x pitch=%d (%s.%d) velocity=%d\n",
				event->start, event->status, event->pitch, 
				note[event->pitch%12], event->pitch/12, event->velocity);
	}
	else if (event->status >= 0xb0 && event->status <= 0xbf) {
		event->pitch = getc(fp);
		event->velocity = getc(fp);
		debug(3, "Parameter: @ %d status=0x%02x param=%d value=%d\n",
				event->start, event->status, event->pitch, event->velocity);
	}
	else if (event->status >= 0xc0 && event->status <= 0xcf) {
		event->pitch = getc(fp);
		debug(1, "Program change: @ %d status=0x%02x program=%d\n",
				event->start, event->status, event->pitch);
	}
	else if (event->status >= 0xd0 && event->status <= 0xdf) {
		event->pitch = getc(fp);
		debug(3, "Channel pressure: @ %d status=0x%02x touch=%d\n",
				event->start, event->status, event->pitch);
	}
	else if (event->status >= 0xe0 && event->status <= 0xef) {
		event->pitch = getc(fp) + (getc(fp) << 7);
		debug(3, "Pitch bend: @ %d status=0x%02x bend=%d\n",
				event->start, event->status, event->pitch);
	}
	else if (event->status == 0xff) {
		event->pitch = getc(fp);
		switch (event->pitch) {
			case 0x00:  /* sequence number: FF 00 02 ss ss _or_ FF 00 00 */
				debug(2, "Ignoring sequence number (0x00) record.\n");
				if (getc(fp) == 0x02) fseek(fp, 2, SEEK_CUR);
				break;
			case 0x01:  /* text: FF 0n len text */
			case 0x02:  /* copyright */
			case 0x03:  /* sequence/track name */
			case 0x04:  /* instrument */
			case 0x05:  /* lyric */
			case 0x06:  /* marker */
			case 0x07:  /* cue point */
			case 0x08:  /* program (patch) name */
			case 0x09:  /* device (port) name */
				len = read_varint(fp);
				event->data = g_new(char, len+1);
				fread(event->data, len, 1, fp);
				event->data[len] = '\0';
				debug(2, "Text (0x%02x) record: \"%s\".\n", event->pitch, event->data);
				break;
			case 0x2f:  /* end of track: FF 2F 00 */
				event->velocity = getc(fp);
				g_assert(event->velocity == 0);
				break;
			case 0x51:  /* tempo: FF 51 03 tt tt tt */
				len = getc(fp);
				g_assert(len == 3);
				event->velocity = (getc(fp)<<16) + (getc(fp)<<8) + getc(fp);
				event->velocity = 60000000/event->velocity;
				debug(3, "Tempo record: %d bpm.\n", event->velocity);
				break;
			case 0x54:  /* SMPTE offset: FF 54 05 hr mn se fr ff */
			case 0x58:  /* Time signature: FF 58 04 nn dd cc bb */
			case 0x59:  /* Key signature: FF 59 02 sf mi */
			case 0x7f:  /* Proprietary: FF 7F len data */
				len = getc(fp);
				fseek(fp, len, SEEK_CUR);
				debug(2, "Ignoring typed event (0x%02x), length %d.\n", event->pitch, len);
				break;
			default:
				len = getc(fp);
				fseek(fp, len, SEEK_CUR);
				debug(2, "Unknown FF event type (0x%02x).  Skipping %d bytes.\n", event->pitch, len);
		}
	}
	else {
		debug(2, "Unknown event type: 0x%02x\n", event->status);
	}
	return event;
}

unsigned short read_int16(FILE *fp) {
	return (getc(fp)<<8) + getc(fp);
}

unsigned long read_int32(FILE *fp) {
	return (getc(fp)<<24) + (getc(fp)<<16) + (getc(fp)<<8) + (getc(fp));
}

unsigned long read_varint(FILE *fp) {
	unsigned long value;
	unsigned char c;

	if ((value = getc(fp)) & 0x80) {
		value &= 0x7F;
		do {
			value = (value << 7) + ((c = getc(fp)) & 0x7F);
		} while (c & 0x80);
	}

	return value;
}

void write_midi_file(FILE *fp, MIDIFile *mf) {
	int i;

	write_header_chunk(fp, mf->header);
	g_assert(mf->header->numtracks == mf->tracks->len);
	for (i=0; i<mf->header->numtracks; i++)
		write_track(fp, TRACK(mf, i));
}

void write_header_chunk(FILE *fp, Header *header) {
	fwrite("MThd", 4, 1, fp);
	write_int32(fp, 6);  /* length */
	write_int16(fp, header->format);
	write_int16(fp, header->numtracks);
	write_int16(fp, header->division);
}

void write_track(FILE *fp, Track *track) {
	long length_pos, end_pos;
	int i;

	fwrite("MTrk", 4, 1, fp);
	length_pos = ftell(fp);
	fwrite("    ", 4, 1, fp);  /* length placeholder */

	for (i=0; i<track->events->len; i++)
		write_event(fp, g_ptr_array_index(track->events, i), i>0 ? g_ptr_array_index(track->events, i-1) : NULL);

	end_pos = ftell(fp);
	fseek(fp, length_pos, SEEK_SET);
	write_int32(fp, end_pos - length_pos - 4);
	fseek(fp, end_pos, SEEK_SET);
}

void write_event(FILE *fp, Event *event, Event *last_event) {
	int delta_time, write_it, len, tempo;

	write_it = (event->status >= 0x80 && event->status <= 0x9f)
		|| (event->status >= 0xb0 && event->status <= 0xef);
	if (!write_it && event->status == 0xff) {
		write_it = (event->pitch >= 0x01 && event->pitch <= 0x09)
			|| event->pitch == 0x2f || event->pitch == 0x51;
	}
	if (!write_it) return;

	delta_time = last_event ? event->start-last_event->start : 0;
	write_varint(fp, delta_time);

	/*!! why doesn't this work? */
	//if (!last_event || last_event->status != event->status)
		putc(event->status, fp);

	if ((event->status >= 0x80 && event->status <= 0x9f)
			|| (event->status >= 0xb0 && event->status <= 0xbf)) {
		putc(event->pitch, fp);
		putc(event->velocity, fp);
	}
	else if (event->status >= 0xc0 && event->status <= 0xdf) {
		putc(event->pitch, fp);
	}
	else if (event->status >= 0xe0 && event->status <= 0xef) {
		putc(event->pitch & 0x7F, fp);
		putc((event->pitch>>7) & 0x7F, fp);
	}
	else if (event->status == 0xff) {
		putc(event->pitch, fp);
		switch (event->pitch) {
			/* 0x00 ignored */
			case 0x01:  /* text: FF 0n len text */
			case 0x02:  /* copyright */
			case 0x03:  /* sequence/track name */
			case 0x04:  /* instrument */
			case 0x05:  /* lyric */
			case 0x06:  /* marker */
			case 0x07:  /* cue point */
			case 0x08:  /* program (patch) name */
			case 0x09:  /* device (port) name */
				len = strlen(event->data);
				write_varint(fp, len);
				fwrite(event->data, len, 1, fp);
				break;
			case 0x2f:  /* end of track: FF 2F 00 */
				putc(0x00, fp);
				g_assert(event->velocity == 0);
				break;
			case 0x51:  /* tempo: FF 51 03 tt tt tt */
				putc(3, fp);
				tempo = 60000000/event->velocity;
				putc((tempo>>16)&0xff, fp);
				putc((tempo>>8)&0xff, fp);
				putc(tempo&0xff, fp);
				break;
			/* 0x54, 0x58, 0x59, 0x7f ignored */
			default:
				g_assert_not_reached();
		}
	}
	else
		g_assert_not_reached();
}

void write_int16(FILE *fp, unsigned short n) {
	putc((n>>8)&0xff, fp);
	putc(n&0xff, fp);
}

void write_int32(FILE *fp, unsigned long n) {
	putc((n>>24)&0xff, fp);
	putc((n>>16)&0xff, fp);
	putc((n>>8)&0xff, fp);
	putc(n&0xff, fp);
}

void write_varint(FILE *fp, unsigned long n) {
	unsigned long buffer;
	buffer = n & 0x7F;

	while ((n >>= 7)) {
		buffer <<= 8;
		buffer |= ((n & 0x7F) | 0x80);
	}

	while (1) {
		putc(buffer, fp);
		if (buffer & 0x80) buffer >>= 8; else break;
	}
}

void merge_tracks(MIDIFile *mf) {
	Event *ev;
	Track *track;
	int i, cur[16];
	track = g_new(Track, 1);
	track->events = g_ptr_array_new();
	bzero(cur, sizeof(cur));
	while (1) {
		int mintime, minidx = -1;
		Track *t;
		for (i=0; i<mf->tracks->len; i++) {
			t = g_ptr_array_index(mf->tracks, i);
			if (cur[i] >= t->events->len) continue;
			ev = g_ptr_array_index(t->events, cur[i]);
			if (minidx == -1 || ev->start < mintime) {
				minidx = i;
				mintime = ev->start;
			}
		}
		if (minidx == -1) break; /* no valid events remaining */
		t = g_ptr_array_index(mf->tracks, minidx);
		ev = g_ptr_array_index(t->events, cur[minidx]);
		if (ev->status != 0xff || ev->pitch != 0x2f)
			/* add all non-end events */
			g_ptr_array_add(track->events, ev);
		else
			free_event(ev);
		cur[minidx]++;
	}

	/* add the end event */
	ev = g_new(Event, 1);
	ev->start = ((Event*)g_ptr_array_index(track->events, track->events->len-1))->start;
	ev->status = 0xff;
	ev->pitch = 0x2f;
	ev->velocity = 0x00;
	ev->data = NULL;
	g_ptr_array_add(track->events, ev);

	for (i=0; i<mf->tracks->len; i++) {
		Track *track = TRACK(mf, i);
		/* can't call free_track, because it kills the events */
		g_ptr_array_free(track->events, TRUE);
		g_free(track);
	}
	g_ptr_array_free(mf->tracks, TRUE);
	mf->tracks = g_ptr_array_new();
	g_ptr_array_add(mf->tracks, track);

	mf->header->numtracks = 1;
}

void free_midi_file(MIDIFile *mf) {
	unsigned int i;
	free_header(mf->header);
	for (i=0; i<mf->tracks->len; i++)
		free_track(g_ptr_array_index(mf->tracks, i));
	g_ptr_array_free(mf->tracks, TRUE);
	g_free(mf);
}

void free_header(Header *header) {
	g_free(header);
}

void free_track(Track *track) {
	unsigned int i;
	for (i=0; i<track->events->len; i++)
		free_event(g_ptr_array_index(track->events, i));
	g_ptr_array_free(track->events, TRUE);
	g_free(track);
}

void free_event(Event *event) {
	if (event->data) g_free(event->data);
	g_free(event);
}

void debug(int level, const char *fmt, ...) {
	va_list arg;

	if (level <= DEBUG_LEVEL) {
		va_start(arg, fmt);
		vfprintf(stderr, fmt, arg);
		va_end(arg);
	}
}
